import random,sys,math
from optparse import OptionParser
from PS9_netsim import *
import PS9_tests

### Skeleton for distance vector routing lab in 6.02

# use our own node class derived from the node class of lab8_net.py
# so we can override routing behavior
class DVRouter(Router):
	INFINITY = 32

	def send_advertisement(self, time):
		adv = self.make_dv_advertisement()
		for link in self.links:
			p = self.network.make_packet(self.address, self.peer(link), 
										'ADVERT', time,
										color='red', ad=adv)
			link.send(self, p)        

	# Make a distance vector protocol advertisement, which will be sent
	# by the caller along all the links
	def make_dv_advertisement(self):
		#collect a list of the spcost and routes
		#print '--'
		#print 'routes'
		#print self.routes
		#print 'costs'
		#print self.spcost
		return self.spcost
		
	def link_failed(self, link):
		#print 'FAILURE on'
		#print link
		dests = self.get_dest_on_link(link)
		for dest in dests:
			#set that link to infinity
			self.spcost[dest] = self.INFINITY
			#delete route
			#must set to None
			self.routes[dest] = None

	def process_advertisement(self, p, link, time):
		self.integrate(link, p.properties['ad'])

	# Integrate new routing advertisement to update routing
	# table and costs
	def integrate(self,link,adv):
		#print 'ad recieved'
		#print 'on link'
		#print link
		#print adv
		for ad in adv:
			dest = ad
			cost = adv[ad]
			newcost = cost + link.cost
			
			#checks
			#first if lower cost path
			if self.get_cost(dest) > newcost:
				#set
				self.routes[dest] = link
				self.spcost[dest] = newcost
			#if same link, update costs
			if self.get_next_link(dest) == link:
				#only update, so readable
				if self.spcost[dest] != newcost:
					#print 'update'
					#print 'old cost'
					#print self.spcost[dest]
					#print 'new cost'
					#print newcost
					self.spcost[dest] = newcost
			#third case: route is now a failure (None) if it came over that same ling
			##dests = get_dest_on_link(link)
			##check every dest if now a None in ad
			##for dest in dests:
			##	#actually we are checking for infinite cost, since we only have costs, not further routes
			##	if 
			## DONT NEED, the failure is propagated by second rule

	def get_cost(self,dest):
		if dest in self.spcost:
			return self.spcost[dest]
		else:
			return self.INFINITY

	def get_next_link(self, dest):
		if dest in self.routes:
			return self.routes[dest]
		else:
			return None

	def get_dest_on_link(self, link):
		dests = []
		for destitem in self.routes:
			#print self.routes[destitem]
			if self.routes[destitem] == link:
				dests.append(destitem)
		return dests

# A network with nodes of type DVRouter.
class DVRouterNetwork(RouterNetwork):
    # nodes should be an instance of DVNode (defined above)
    def make_node(self,loc,address=None):
        return DVRouter(loc,address=address)

########################################################################

if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option("-g", "--gui", action="store_true", dest="gui", 
                      default=False, help="show GUI")
    parser.add_option("-n", "--numnodes", type="int", dest="numnodes", 
                      default=12, help="number of nodes")
    parser.add_option("-t", "--simtime", type="int", dest="simtime", 
                      default=2000, help="simulation time")
    parser.add_option("-r", "--rand", action="store_true", dest="rand", 
                      default=False, help="use randomly generated topology")
#    parser.add_option("-l", "--loss", type="float", dest="lossprob", 
#                      default=0.0, help="link packet loss probability")
#    parser.add_option("-f", "--mttf", type="int", dest="mttf", 
#                      default=10000, help="mean time between failures")
    
    (opt, args) = parser.parse_args()

    if opt.rand == True:
        rg = RandomGraph(opt.numnodes)
        (NODES, LINKS) = rg.genGraph()
    else:
        # build the deterministic test network
        #   A---B   C---D
        #   |   | / | / |
        #   E   F---G---H
        # format: (name of node, x coord, y coord)

        NODES =(('A',0,0), ('B',1,0), ('C',2,0), ('D',3,0),
                ('E',0,1), ('F',1,1), ('G',2,1), ('H',3,1))

        # format: (link start, link end)
        LINKS = (('A','B'),('A','E'),('B','F'),('E','F'),
                 ('C','D'),('C','F'),('C','G'),
                 ('D','G'),('D','H'),('F','G'),('G','H'))

    # setup graphical simulation interface
    if opt.gui == True:
        net = DVRouterNetwork(opt.simtime, NODES, LINKS, 0)
        sim = NetSim()
        sim.SetNetwork(net)
        sim.MainLoop()
    else:
        PS9_tests.verify_routes(DVRouterNetwork)
